/* -*-C-*-
*******************************************************************************
*
* File:         hpl_tiff.c
* RCS:          $Header: hpl_tiff.c,v 1.6 95/05/22 20:41:48 hmgr Exp $
* Description:  TIFF format handler extension to hplsh.  Uses calls
*               from the freely available TIFF 6.0 package, so support
*               should be fairly complete.
* Author:       Andrew Fitzhugh
* Created:      Wed May  3 09:55:28 1995
* Modified:     Mon May 22 20:30:58 1995 (AEF) fitz@hplaef.hpl.hp.com
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 1995, Hewlett-Packard Laboratories, all rights reserved.
*
*******************************************************************************
*/

#include <tcl.h>
#include <tk.h>
#include <tiffio.h>
#include <fcntl.h>

static int		FileMatchTIFF _ANSI_ARGS_((FILE *f, char *fileName,
			    char *formatString, int *widthPtr,
			    int *heightPtr));
static int		FileReadTIFF  _ANSI_ARGS_((Tcl_Interp *interp,
			    FILE *f, char *fileName, char *formatString,
			    Tk_PhotoHandle imageHandle, int destX, int destY,
			    int width, int height, int srcX, int srcY));
static int		FileWriteTIFF _ANSI_ARGS_((Tcl_Interp *interp,
			    char *fileName, char *formatString,
			    Tk_PhotoImageBlock *blockPtr));

/*
 * The maximum amount of memory to allocate for data read from the
 * file.  If we need more than this, we do it in pieces.
 */

#define MAX_MEMORY	10000	/* don't allocate > 10KB */

Tk_PhotoImageFormat tkImgFmtTIFF = {
    "TIFF",			/* name */
    FileMatchTIFF,		/* fileMatchProc */
    NULL,			/* stringMatchProc */
    FileReadTIFF,		/* fileReadProc */
    NULL,			/* stringReadProc */
    FileWriteTIFF,	        /* fileWriteProc */
    NULL,			/* stringWriteProc */
};

/*
 * Prototypes for local procedures defined in this file:
 */

static TIFF            *ReadTIFFHeader _ANSI_ARGS_((FILE *f, char *fileName,
			    int *widthPtr, int *heightPtr));


/*
 *----------------------------------------------------------------------
 *
 * FileMatchTIFF --
 *
 *	This procedure is invoked by the photo image type to see if
 *	a file contains image data in TIFF format.
 *
 * Results:
 *	True if the TIFF magic number is found in a valid TIFF header,
 *      otherwise False.
 *
 * Side effects:
 *      The TIFF library doesn't have an existing routine to only read
 *      the width & height from the image;  I simply allocate & read the
 *      whole TIFF directory, then de-allocate it.  This will likely
 *      impact performance and should be replaced by a routine that
 *      only extracts the width & height (when I have time!).
 *
 *      Note that f is not used.  The tiff library routines do not work
 *      well from an open stream (there is no way to deallocate the TIFF
 *      structure without closing the stream).  The fileName variable
 *      is used, and a new file stream is opened for it.
 *
 *----------------------------------------------------------------------
 */

static int
FileMatchTIFF(f, fileName, formatString, widthPtr, heightPtr)
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here if the file is a valid
				 * TIFF file. */
{
    TIFF *tif;

    tif = ReadTIFFHeader(f, fileName, widthPtr, heightPtr);
    if (tif) {
	TIFFClose(tif);
	return(1);
    }
    else {
	return(0);
    }
}




/*
 *----------------------------------------------------------------------
 *
 * TIFFPromoteTo24 --
 *
 *	This procedure converts a scanline of TIFF data from any
 *      valid TIFF format to 24 bits/pixel.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *	
 * Side effects:
 *      Discards any alpha channel pixels.  Each channel is converted
 *      from N bits (1,4,8) to 8 bits per sample, and from M samples
 *      to 3 samples per pixel.
 *
 *----------------------------------------------------------------------
 */

static int
checkcmap(int n, unsigned short *r, unsigned short *g, unsigned short *b)
{
    while (n-- > 0)
	if (*r++ >= 256 || *g++ >= 256 || *b++ >= 256)
	    return (16);
    fprintf(stderr, "Warning, assuming 8-bit colormap.\n");
    return (8);
}

/* 
 * This macro decodes bit-packed formats into bytes.  Supported
 * BitsPerPixel values are 1, 4, and 8, as per the TIFF 6.0 spec.
 * Yes, this really should be a macro, for various reasons.
 */

#define getBits(in, out) { \
    if ( bitsleft == 0 ) { \
        ++(in); \
        bitsleft = 8; \
    } \
    bitsleft -= bps; \
    (out) = ( *(in) >> bitsleft ) & maxval; \
    if (maxval != 255) (out) = (out) * 255 / maxval; }

static int TIFFPromoteTo24(src, dst, num, bps, spp, photomet,
			   red, green, blue, maxval)
unsigned char *src, *dst;
int num;
unsigned short bps, spp, photomet;
unsigned short *red, *green, *blue;
unsigned char maxval;
/* TIFF *tif; */
{
    int i, j;
    int bitsleft = 8;
    unsigned char tmp, *inP;

    /*
     * Promote all pixel formats to 24 bits/pixel
     */
    inP = src;
    for (j=0; j<num; j++) {
	switch (photomet) {
	case PHOTOMETRIC_MINISBLACK:
	    getBits(inP, tmp);
	    *dst++ = tmp;
	    *dst++ = tmp;
	    *dst++ = tmp;
	    
	    /* Skip alpha channels */
	    for (i=1; i<spp; i++) getBits(inP, tmp);
	    break;
	case PHOTOMETRIC_MINISWHITE:
	    getBits(inP, tmp);
	    tmp = ~tmp;
	    *dst++ = tmp;
	    *dst++ = tmp;
	    *dst++ = tmp;
	    
	    /* Skip alpha channels */
	    for (i=1; i<spp; i++) getBits(inP, tmp); 
	    break;
	case PHOTOMETRIC_PALETTE:
	    getBits(inP, tmp);
	    *dst++ = red[tmp];
	    *dst++ = green[tmp];
	    *dst++ = blue[tmp];
	    
	    /* Skip alpha channels */
	    for (i=1; i<spp; i++) getBits(inP, tmp);
	    break;
	case PHOTOMETRIC_RGB:
	    getBits(inP, tmp);
	    *dst++ = tmp;
	    getBits(inP, tmp);
	    *dst++ = tmp;
	    getBits(inP, tmp);
	    *dst++ = tmp;
	    
	    /* Skip alpha channels */
	    for (i=3; i<spp; i++) getBits(inP, tmp);
	    break;
	default:
	    fprintf(stderr,
		    "TIFFPromoteTo24() error: Should not get here!!\n");
	    break;
	}
    }
    return(0);
}


/*
 *----------------------------------------------------------------------
 *
 * FileReadTIFF --
 *
 *	This procedure is called by the photo image type to read
 *	TIFF format data from a file and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *	
 * Side effects:
 *      The file stream pointer, f, is not used.  The fileName is
 *      passed to the tiff library, and a new stream is opened & closed.
 *
 *      All pixel formats are promoted to 24-bit color.
 *
 *----------------------------------------------------------------------
 */

static int
FileReadTIFF(interp, f, fileName, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
    int fileWidth, fileHeight;
    int nLines, nBytes, h, i;
    unsigned char *pixelPtr;
    Tk_PhotoImageBlock block;
    int dstx, dsty;

    /* TIFF variables */
    TIFF *tif;
    unsigned char *tif_rowbuf;
    int tif_row;
    int e;
    unsigned short tif_bps, tif_spp, tif_photomet;
    unsigned short *rmap;
    unsigned short *gmap;
    unsigned short *bmap;
    unsigned char tif_maxval;
    int cmap_bits;
    

    if ((tif=ReadTIFFHeader(f, fileName, &fileWidth, &fileHeight)) == NULL) {
	Tcl_AppendResult(interp, "couldn't read TIFF header from file \"",
		fileName, "\"", NULL);
	return TCL_ERROR;
    }
    if ((fileWidth <= 0) || (fileHeight <= 0)) {
	Tcl_AppendResult(interp, "TIFF image file \"", fileName,
		"\" has dimension(s) <= 0", (char *) NULL);
	return TCL_ERROR;
    }


    /*
     * Read TIFF parameters which we will use later when promoting
     * all data formats to 24 bits/pixel.
     */
    if ( ! TIFFGetField( tif, TIFFTAG_BITSPERSAMPLE, &tif_bps ) ) {
	tif_bps = 1;
    }
    if ( ! TIFFGetField( tif, TIFFTAG_SAMPLESPERPIXEL, &tif_spp ) ) {
	tif_spp = 1;
    }
    if ( ! TIFFGetField( tif, TIFFTAG_PHOTOMETRIC, &tif_photomet ) ) {
	Tcl_AppendResult(interp, "Could not find photometric tag", 
			 (char *) NULL);
	return TCL_ERROR;
    }
    
    tif_maxval = ( 1 << tif_bps ) - 1;
    
    switch ( tif_photomet ) {
    case PHOTOMETRIC_MINISBLACK:
    case PHOTOMETRIC_MINISWHITE:
    case PHOTOMETRIC_RGB:
	/* OK, can handle these fine. */
	break;
    case PHOTOMETRIC_PALETTE:
	if (!TIFFGetField(tif, TIFFTAG_COLORMAP, 
			  &rmap, &gmap, &bmap)) {
	    Tcl_AppendResult(interp, "Missing required \"Colormap\" tag",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	cmap_bits = checkcmap(1<<tif_bps, rmap, gmap, bmap);
        if (cmap_bits == 16) {
	    /*
	     * Convert 16-bit colormap to 8-bit.  8-bit colormap
	     * is NOT LEGAL TIFF, but it is a common error so let's
	     * deal with it.
	     */
#define CVT(x)          (((x) * 255) / ((1L<<16)-1))
	    for (i = (1<<tif_bps)-1; i > 0; i--) {
		rmap[i] = CVT(rmap[i]);
		gmap[i] = CVT(gmap[i]);
		bmap[i] = CVT(bmap[i]);
                }
        }

	break;                
    case PHOTOMETRIC_MASK:
    default:
	Tcl_AppendResult(interp, "Unknown photometric type", (char *) NULL);
	return TCL_ERROR;
	break;
    }


    if ((srcX + width) > fileWidth) {
	width = fileWidth - srcX;
    }
    if ((srcY + height) > fileHeight) {
	height = fileHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
	|| (srcX >= fileWidth) || (srcY >= fileHeight)) {
	return TCL_OK;
    }

    block.width = width;
    block.pixelSize = 3;
    block.pitch = 3 * fileWidth;
    block.offset[0] = 0;
    block.offset[1] = 1;
    block.offset[2] = 2;

    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

    if (TIFFIsTiled(tif)) {
	unsigned int row, col, tilew;
	unsigned int tl, tw;
	u_char *bufp;
	
	if ((tif_rowbuf = (u_char *) _TIFFmalloc(TIFFTileSize(tif))) == NULL) {
	    fprintf(stderr,"Can't allocate memory for TIFF tile buffer\n");
	    return TCL_ERROR;
	}
	
	tilew = TIFFTileRowSize(tif);
	TIFFGetField(tif, TIFFTAG_TILEWIDTH, &tw);
	TIFFGetField(tif, TIFFTAG_TILELENGTH, &tl);
	
	block.width = tw;
	block.pixelSize = 3;
	block.pitch = 3 * tw;
	block.offset[0] = 0;
	block.offset[1] = 1;
	block.offset[2] = 2;

	nBytes = tl * block.pitch;
	pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
	block.pixelPtr = pixelPtr;
	
	dsty = destY;

	for (row = 0; row < height; row += tl) {
	    dstx = destX;

	    if (row + tl > height) block.height = height -row;
	    else block.height = tl;

	    bufp = pixelPtr;
	    for (col = 0; col < width; col += tw) {

		if (TIFFReadTile(tif, tif_rowbuf, col, row, 0, 0) < 0) {
		    Tcl_AppendResult(interp, "error reading TIFF tile \"",
				     fileName, "\": ", Tcl_PosixError(interp),
				     (char *) NULL);
		    TIFFClose(tif);
		    ckfree((char *) pixelPtr);
		    _TIFFfree(tif_rowbuf);
		    return TCL_ERROR;
		}

		/*
		 * Convert the TIFF tile to a photo widget tile (basically,
		 * this means promoting all tiff data to 24-bit RGB).
		 */
		for (i=0; i<block.height; i++) {
		    TIFFPromoteTo24(tif_rowbuf+i*tilew, bufp+i*3*tw, 
				    block.width, tif_bps, tif_spp, 
				    tif_photomet, rmap, gmap, bmap,
				    tif_maxval);
		}

		if (col + tw > width) block.width = width - col;
		else block.width = tw;

		Tk_PhotoPutBlock(imageHandle, &block, dstx, dsty, 
				 block.width, block.height);
		dstx += tw;
	    }
	    dsty += tl;
	}
    }
    else {
	nLines = (MAX_MEMORY + block.pitch - 1) / block.pitch;
	if (nLines > height) {
	    nLines = height;
	}
	if (nLines <= 0) {
	    nLines = 1;
	}

	nBytes = nLines * block.pitch;
	pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
	block.pixelPtr = pixelPtr + srcX * 3;
	
	if ((tif_rowbuf=(unsigned char*) _TIFFmalloc(TIFFScanlineSize(tif))) ==
	    NULL) {
	    fprintf(stderr,"Can't allocate memory for TIFF scanline buffer\n");
	    return TCL_ERROR;
	}
	
	tif_row = srcY;
	for (h = height; h > 0; h -= nLines) {
	    if (nLines > h) {
		nLines = h;
		nBytes = nLines * block.pitch;
	    }
	    
	    for (i=0; i<nLines; i++) {
		e = TIFFReadScanline(tif, tif_rowbuf, tif_row, 0);
		if (e < 0) {
		    Tcl_AppendResult(interp, "error reading TIFF image file \"",
				     fileName, "\": ", Tcl_PosixError(interp),
				     (char *) NULL);
		    TIFFClose(tif);
		    ckfree((char *) pixelPtr);
		    _TIFFfree(tif_rowbuf);
		    return TCL_ERROR;
		}
		
		e = TIFFPromoteTo24(tif_rowbuf, pixelPtr+i*fileWidth*3, 
				    fileWidth, tif_bps, tif_spp, tif_photomet, 
				    rmap, gmap, bmap, tif_maxval);
		
		if (e < 0) {
		    Tcl_AppendResult(interp, "error promoting TIFF data \"",
				     fileName, "\": ", Tcl_PosixError(interp),
				     (char *) NULL);
		    TIFFClose(tif);
		    ckfree((char *) pixelPtr);
		    free(tif_rowbuf);
		    return TCL_ERROR;
		}
		
		tif_row++;
	    }
	    
	    block.height = nLines;
	    Tk_PhotoPutBlock(imageHandle, &block, destX, destY, width, nLines);
	    destY += nLines;
	}
    }

    TIFFClose(tif);
    ckfree((char *) pixelPtr);
    _TIFFfree(tif_rowbuf);

    return TCL_OK;
}



/*
 *----------------------------------------------------------------------
 *
 * FileWriteTIFF --
 *
 *	This procedure is invoked to write image data to a file in TIFF
 *	format.  The image type will be 8 bits/sample, 3 samples/pixel,
 *      with no compression.  It would be very easy to write compressed
 *      data, but the photo widget has no way to communicate such an
 *      option to the FileWrite procedure.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	Data is written to the file given by "fileName".
 *
 *----------------------------------------------------------------------
 */

static int
FileWriteTIFF(interp, fileName, formatString, blockPtr)
    Tcl_Interp *interp;
    char *fileName;
    char *formatString;
    Tk_PhotoImageBlock *blockPtr;
{
    int w, h, row, col;
    int greenOffset, blueOffset, nBytes;
    unsigned char *pixelPtr, *pixLinePtr;
    unsigned long rowsperstrip = (unsigned long) -1;
    unsigned long linebytes;
    unsigned char *row_buf, *rp;
    TIFF *out;


    out = TIFFOpen(fileName, "wb");
    if ((out = TIFFOpen(fileName, "wb")) == NULL) {
	Tcl_AppendResult(interp, fileName, ": ", Tcl_PosixError(interp),
		(char *)NULL);
	return TCL_ERROR;
    }
    w = blockPtr->width;
    h = blockPtr->height;

    fprintf(stderr,"Writing tiff file\n");

    /*
     * Set up the TIFF header for 8 bits/sample, 3 samples/pixel RGB.
     */
    TIFFSetField(out, TIFFTAG_IMAGEWIDTH,  w);
    TIFFSetField(out, TIFFTAG_IMAGELENGTH, h);
    TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 3);
    TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    TIFFSetField(out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    linebytes = 3 * w;
    if (rowsperstrip == (unsigned long) -1)
	rowsperstrip = (8*1024)/linebytes;
    if (rowsperstrip == 0)
	rowsperstrip = 1;
    TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, rowsperstrip);


    pixLinePtr = blockPtr->pixelPtr + blockPtr->offset[0];
    greenOffset = blockPtr->offset[1] - blockPtr->offset[0];
    blueOffset = blockPtr->offset[2] - blockPtr->offset[0];

    fprintf(stderr,"Set fields\n");

    if ((greenOffset == 1) && (blueOffset == 2) && (blockPtr->pixelSize == 3)
	&& (blockPtr->pitch == (blockPtr->width * 3))) {
	/*
	 * We can optimize this common case, so do it.
	 */
	for (row = 0; row<h; row++) {
	    fprintf(stderr,"line %d\n",row);
	    if (TIFFWriteScanline(out, pixLinePtr, row, 0) < 0)
		goto writeerror;
	    pixLinePtr += w * 3;
	}
    } else {
	row_buf = (unsigned char *) ckalloc((unsigned) w*3);

	for (row=0; row < h; row++) {
	    rp = row_buf;
	    pixelPtr = pixLinePtr;
	    for (col=0; col < w; col++) {
		*rp++ = pixelPtr[0];
		*rp++ = pixelPtr[greenOffset];
		*rp++ = pixelPtr[blueOffset];
		pixelPtr += blockPtr->pixelSize;
	    }
	    if (TIFFWriteScanline(out, row_buf, row, 0) < 0)
		goto writeerror;
	    pixLinePtr += blockPtr->pitch;
	}
    }

    TIFFClose(out);
    ckfree((char *) row_buf);
    return TCL_OK;

 writeerror:
    Tcl_AppendResult(interp, "error writing \"", fileName, "\": ",
	    Tcl_PosixError(interp), (char *) NULL);
    TIFFClose(out);
    ckfree((char *) row_buf);
    return TCL_ERROR;
}



/*
 *----------------------------------------------------------------------
 *
 * ReadTIFFHeader --
 *
 *	This procedure reads the TIFF header and reads the 
 *      directory information into the TIFF data structure.
 *
 * Results:
 *
 * Side effects:
 *	The access position in f advances.
 *
 *----------------------------------------------------------------------
 */

static TIFF
*ReadTIFFHeader(f, fileName, widthPtr, heightPtr)
    FILE *f;			/* Image file to read the header from */
    char *fileName;             /* Name of the file */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here. */
{
    TIFF *tif = NULL;
    char ch1, ch2;

    ch1 = fgetc(f);
    ch2 = fgetc(f);
    if ((ch1 == 'M' && ch2 == 'M') || (ch1 == 'I' && ch2 == 'I')) {
/*	if (!(tif = TIFFFdOpen(fileno(f), fileName, "r"))) { */
	TIFFSetWarningHandler(NULL);
	if (!(tif = TIFFOpen(fileName, "r"))) {
	    fprintf(stderr,"Put TCL warning here -- TIFF open failed\n");
	    return(0);
	}
	TIFFGetField( tif, TIFFTAG_IMAGEWIDTH, widthPtr );
        TIFFGetField( tif, TIFFTAG_IMAGELENGTH, heightPtr );
    }

    return(tif);
}



/*
 *----------------------------------------------------------------------
 *
 * Tiff_Init --
 *
 *	This procedure is typically invoked by Tcl_AppInit procedures
 *	to perform additional Tiff initialization.
 *
 * Results:
 *	Returns a standard Tcl completion code and sets interp->result
 *	if there is an error.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
Tiff_Init(interp)
    Tcl_Interp *interp;		/* Interpreter to initialize. */
{
    Tk_CreatePhotoImageFormat(&tkImgFmtTIFF);

    return TCL_OK;
}

